/* f.c - implementation of the f interpreter as a command line program */
/* Copyright 2012 Miller Puckette - standard BSD license */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/select.h>

#define STACKSIZE 1000
#define MEMSIZE 10

#define E_FLOAT 0       /* floating-point number (for e_type field below) */
#define E_FUNCTION 1    /* built-in function.. */
#define E_OPEN 2        /* open parenthesis */
#define E_CLOSE 3       /* close parenthesis */

struct _interp;     /* an as-yet unspecified struct we need for next line */

typedef int (*t_builtin)(struct _interp *x);    /* type for built-in fns */

typedef struct _elem    /* an element. Check type before using union */
{
    int e_type; /* float, function, open or close paren as above */
    union       /* a word holding whatever the "type" specifies */
    {
        float e_float;      /* only valid if e_elem is E_FLOAT */
        t_builtin e_fn;     /* or E_FUNCTION.  Undefined for E_OPEN, E_CLOSE */
    } ;
} t_elem;

#define RTN_DONE  0       /* returned when program finished */
#define RTN_ERROR 1       /* returned because of an error */
#define RTN_INPUT 2       /* returned to ask for input */
#define RTN_WAIT  3       /* returned to ask for sleep time */
#define RTN_OUTPUT  4     /* returned to make output */

typedef struct _interp      /* state of an interpreter: memory and a stack */
{
    t_elem i_stack[STACKSIZE];  /* the stack is a static array of elems */
    int i_stacksize;            /* number of elems currently on the stack */
    int i_stackframe;           /* point not to pop past in current context */
    int i_pc;                   /* program counter (index into list to eval) */
    int i_endblk;               /* index of end list to eval */
    struct
    {
        int m_size;             /* size of list in this memory cell */
        t_elem *m_stuff;        /* contents of this cell (dynamic storage) */
    } i_memory[MEMSIZE];        /* fixed number of lists for 'memory' */
} t_interp;

char *fntostring(t_builtin fn); /* we'll use this before we can define it */
    /* BAIL uses a twist in C syntax to allow us to "return()" in a macro */
#define BAIL(x) do {fprintf(stderr, "%s\n", x); return (1);} while (0)

int pushelem(t_interp *x, t_elem *a)    /* push an elem onto the stack */
{
    if (x->i_stacksize >= STACKSIZE)
        BAIL("stack overflow");         /* return 1 on error */
    x->i_stack[x->i_stacksize] = *a;
    x->i_stacksize++;
    return (0);                         /* return 0 on success */
}

int pushnumber(t_interp *x, float f)    /* push a number onto the stack */
{
    if (x->i_stacksize >= STACKSIZE)
        BAIL("stack overflow");
    x->i_stack[x->i_stacksize].e_type = E_FLOAT;
    x->i_stack[x->i_stacksize].e_float = f;
    x->i_stacksize++;
    return (0);
}

int popnumber(t_interp *x, float *f)    /* pop a number; flag error if we */
{                                       /* get anything but a number */
    if (x->i_stacksize <= x->i_stackframe ||
        x->i_stack[x->i_stacksize-1].e_type != E_FLOAT)
            return (1);
    *f = x->i_stack[--x->i_stacksize].e_float;
    return (0);
}

int poplist(t_interp *x, t_elem **block, int *blocksize) /* pop a list inside */
{                                                        /* parens or an elem */
    int parendepth = 0, i;                               /* if no parens */
    if (x->i_stacksize <= x->i_stackframe)
        return (1);
    if (x->i_stack[x->i_stacksize-1].e_type != E_CLOSE)
    {
        *block = &x->i_stack[x->i_stacksize-1];     /* just return one elem */
        *blocksize = 1;
        x->i_stacksize--;
        return (0);
    }
    for (i = x->i_stacksize; i-- > x->i_stackframe; ) /* find matching '('  */
    {                                /* and return list between them but not */
        if (x->i_stack[i].e_type == E_CLOSE)        /* the parens themselves */
            parendepth++;
        else if (x->i_stack[i].e_type == E_OPEN)
            parendepth--;
        if (parendepth == 0)
        {
            *block = &x->i_stack[i+1]; 
            *blocksize = x->i_stacksize - i - 2;  
            x->i_stacksize = i;
            return (0);
        }
    }
    return (1);     /* close paren at stack head was never opened */
}

void interp_start(t_interp *x, int blocksize)   /* start a new exec sequence */
{
    x->i_stackframe = x->i_pc = 0;  /* not quite right yet  */
    x->i_endblk = blocksize;
}

void interp_init(t_interp *x)   /* initialize a new interpreter */
{
    int i;
    for (i = 0; i < MEMSIZE; i++)
        x->i_memory[i].m_size = 0, x->i_memory[i].m_stuff = 0;
    interp_start(x, 0);
}

void interp_clear(t_interp *x)   /* clear everything */
{
    int i;
    for (i = 0; i < MEMSIZE; i++)
    {
        x->i_memory[i].m_size = 0;
        x->i_memory[i].m_stuff = realloc(x->i_memory[i].m_stuff, 0);
    }
    interp_start(x, 0);
}

void interp_free(t_interp *x)   /* free memory used by the interpreter */
{
    int i;
    for (i = 0; i < MEMSIZE; i++)
        free(x->i_memory[i].m_stuff);
}


int interp_eval(t_interp *x, t_elem *block)     /* call this repeatedly */
{                                           /*  until it returns with "done" */
    int parendepth = 0, ret;
    t_elem *wherecode = (x->i_stackframe ? x->i_stack : block);
    t_elem *nexte = wherecode + x->i_pc, *laste = wherecode + x->i_endblk;
    while (1)
    {
        if (nexte == laste)     /* end of the block? */
        {
            x->i_pc = nexte - wherecode;
            if (x->i_stackframe)
            {
                (*(x->i_stack[x->i_stackframe-1].e_fn))(x);
                wherecode = (x->i_stackframe ? x->i_stack : block);
                nexte = wherecode + x->i_pc;
                laste = wherecode + x->i_endblk;
            }
            else return (RTN_DONE);
        }
        else
        {
            if (nexte->e_type == E_OPEN)  /* maintain paren depth */
                parendepth++;
            if (parendepth)                 /* if inside parens, just push */
            {
                if (pushelem(x, nexte))
                    return (RTN_ERROR);
            }
            else if (nexte->e_type == E_FLOAT)    /* number: push on stack */
            {
                if (pushnumber(x, nexte->e_float))
                    return (RTN_ERROR);
            }
            else if (nexte->e_type == E_FUNCTION)
            {
                x->i_pc = nexte - wherecode;
                if (ret = (*(nexte->e_fn))(x)) /* call a builtin function */
                {
                    x->i_pc++;
                    return (ret);
                }
                wherecode = (x->i_stackframe ? x->i_stack : block);
                nexte = wherecode + x->i_pc;    /* function we called might */
                laste = wherecode + x->i_endblk;    /* have changed PC */
            }
            if (nexte->e_type == E_CLOSE)     /* close paren drops depth */
            {
                if (!parendepth)
                    BAIL("extra close parenthesis");
                parendepth--;
            }
            nexte++;
        }
    }
    return (0);
}

    /****************** BUILT-IN FUNCTIONS START HERE **************/
int f_if_done(t_interp *x)  /* called when an "if" block finishes; pop a */
{                           /* stack frame to resume "calling" block */
    t_elem *block;
    int blocksize;
    x->i_stacksize = x->i_stackframe - 4;
    x->i_pc = x->i_stack[x->i_stackframe-3].e_float;
    x->i_endblk = x->i_stack[x->i_stackframe-2].e_float;
    x->i_stackframe = x->i_stack[x->i_stackframe-4].e_float;
    if (poplist(x, &block, &blocksize))
        BAIL("bad stack frame");
    return (0);
}

int f_if(t_interp *x)   /* test a number and if nonzero, execute a list */
{
    float f;
    t_elem *block;
    int blocksize, stackwas = x->i_stacksize;
    if (popnumber(x, &f)                    /* pop the condition */
        || poplist(x, &block, &blocksize))  /* pop the list to eval */
            BAIL("'if': bad arguments");
    if (f != 0)     /* if the test succeeds, execute the block */
    {
        int newstack = x->i_stacksize;      /* un-pop the list but not the */
        x->i_stacksize = stackwas - 1;      /* number we tested */
        if (x->i_stacksize + 4 >= STACKSIZE)
            BAIL("stack overflow");
        pushnumber(x, x->i_stackframe);
        pushnumber(x, x->i_pc+1);
        pushnumber(x, x->i_endblk);
        x->i_stack[x->i_stacksize].e_type = E_FUNCTION;
        x->i_stack[x->i_stacksize].e_fn = f_if_done;
        x->i_stacksize++;
        x->i_pc = block - x->i_stack - 1;
        x->i_endblk = x->i_pc + blocksize + 1;
        x->i_stackframe = x->i_stacksize;
    }
    return (0);
}

int f_while_body_done(t_interp *x);

int f_while_test_done(t_interp *x)  /* end of test block for f_while below */
{
    float result;
    if (popnumber(x, &result))          /* get numerical result */
        BAIL("'while': arg 1 didn't return a number");
    if (x->i_stackframe < 8 ||    /* sanity check - shouldn't need this */
        x->i_stack[x->i_stackframe-8].e_type != E_FLOAT || 
        x->i_stack[x->i_stackframe-7].e_type != E_FLOAT || 
        x->i_stack[x->i_stackframe-6].e_type != E_FLOAT || 
        x->i_stack[x->i_stackframe-5].e_type != E_FLOAT || 
        x->i_stack[x->i_stackframe-4].e_type != E_FLOAT || 
        x->i_stack[x->i_stackframe-3].e_type != E_FLOAT || 
        x->i_stack[x->i_stackframe-2].e_type != E_FLOAT)
            BAIL("bad stack frame");
    if (result == 1)                    /* true: proceed to body */
    {
        x->i_pc = x->i_stack[x->i_stackframe-3].e_float;
        x->i_endblk = x->i_pc + x->i_stack[x->i_stackframe-2].e_float;
        x->i_stack[x->i_stackframe-1].e_fn = f_while_body_done;
    }
    else    /* false - pop stack frame and restore calling pc and endblk */
    {
        t_elem *block;
        int blocksize;
        x->i_stacksize = x->i_stackframe - 8;
        x->i_pc = x->i_stack[x->i_stackframe-7].e_float;
        x->i_endblk = x->i_stack[x->i_stackframe-6].e_float;
        x->i_stackframe = x->i_stack[x->i_stackframe-8].e_float;
        if (poplist(x, &block, &blocksize) || poplist(x, &block, &blocksize))
            BAIL("bad stack frame");
    }
    return (0);
}

int f_while_body_done(t_interp *x)  /* end of body block for f_while below */
{
    if (x->i_stackframe < 8 ||    /* sanity check - shouldn't need this */
        x->i_stack[x->i_stackframe-8].e_type != E_FLOAT || 
        x->i_stack[x->i_stackframe-7].e_type != E_FLOAT || 
        x->i_stack[x->i_stackframe-6].e_type != E_FLOAT || 
        x->i_stack[x->i_stackframe-5].e_type != E_FLOAT || 
        x->i_stack[x->i_stackframe-4].e_type != E_FLOAT || 
        x->i_stack[x->i_stackframe-3].e_type != E_FLOAT || 
        x->i_stack[x->i_stackframe-2].e_type != E_FLOAT)
            BAIL("bad stack frame");
    x->i_pc = x->i_stack[x->i_stackframe-5].e_float;
    x->i_endblk = x->i_pc + x->i_stack[x->i_stackframe-4].e_float;
    x->i_stacksize = x->i_stackframe;
    x->i_stack[x->i_stackframe-1].e_fn = f_while_test_done;
    return (0);
}

int f_while(t_interp *x)  /* A B while: while list B returns true, exec A too */
{
    t_elem *block1, *block2;
    int blocksize1, blocksize2, stackwas = x->i_stacksize, newstack;
    if (poplist(x, &block1, &blocksize1)
        || poplist(x, &block2, &blocksize2))
            BAIL("'while': bad arguments");
    newstack = x->i_stacksize;  /* un-pop the two lists */
    x->i_stacksize = stackwas;
    if (x->i_stacksize + 8 >= STACKSIZE)
        BAIL("stack overflow");
    pushnumber(x, x->i_stackframe);
    pushnumber(x, x->i_pc+1);
    pushnumber(x, x->i_endblk);
    pushnumber(x, block1 - x->i_stack);
    pushnumber(x, blocksize1);
    pushnumber(x, block2 - x->i_stack);
    pushnumber(x, blocksize2);
    x->i_stack[x->i_stacksize].e_type = E_FUNCTION;
    x->i_stack[x->i_stacksize].e_fn = f_while_test_done;
    x->i_stacksize++;
    x->i_pc = block1 - x->i_stack - 1;
    x->i_endblk = x->i_pc + blocksize1 + 1;
    x->i_stackframe = x->i_stacksize;
    return (0);
}

int f_read(t_interp *x) /* read a list from memory onto stack */
{                       /* if more than one item, put parens around it. */
    float findx;        /* (if empty, also put parens around it; so popping */
    int indx, i;        /* will reliably get back to the previous stack */
    t_elem a;
    if (popnumber(x, &findx))
        BAIL("'read': bad argument");
    if ((indx = findx) < 0 || indx >= MEMSIZE)
        BAIL("'read': variable number out of range");
    if (x->i_memory[indx].m_size == 1)
        pushelem(x, &x->i_memory[indx].m_stuff[0]);
    else
    {
        a.e_type = E_OPEN;
        pushelem(x, &a);
        for (i = 0; i < x->i_memory[indx].m_size; i++)
            pushelem(x, &x->i_memory[indx].m_stuff[i]);
        a.e_type = E_CLOSE;
        pushelem(x, &a);
    }
    return (0);
}

int f_read1(t_interp *x)    /* read one elem of a list from mem onto stack */
{
    float findx, findx2;    /* "F G r" reads Gth item of the list found */
    int indx, indx2, i;     /* at memory location F */
    t_elem a;
    if (popnumber(x, &findx2) || popnumber(x, &findx))
        BAIL("'read1': bad arguments");
    if ((indx = findx) < 0 || indx >= MEMSIZE ||
        (indx2 = findx2) < 0 || indx2 >=  x->i_memory[indx].m_size)
            BAIL("'read1': variable number or index out of range");
    pushelem(x, &x->i_memory[indx].m_stuff[indx2]);
    return (0);
}

int f_write(t_interp *x)    /* 'A F w' pops a list A off the stack and */
{                           /*  writes it to memory location F. */
    float findx, fval;
    int indx, blocksize;
    t_elem *block;
    if (popnumber(x, &findx))                   /* pop F first */
        BAIL("'write': bad arguments");
    if ((indx = findx) < 0 || indx >= MEMSIZE)
        BAIL("'write': variable number out of range");
    if (poplist(x, &block, &blocksize))      /* then A */
        BAIL("'write': bad arguments");
    if (x->i_memory[indx].m_size != blocksize)  /* resize the memory element */
        x->i_memory[indx].m_stuff =
            (t_elem *)realloc(x->i_memory[indx].m_stuff,
                blocksize * sizeof(t_elem));
    x->i_memory[indx].m_size = blocksize;       /* copy A to it */
    memcpy(x->i_memory[indx].m_stuff, block, blocksize*sizeof(t_elem));
    return (0);
}

int f_write1(t_interp *x)   /* 'F G H w1' overwrites the Hth element of the */
{                           /* list at memory location G with the value F. */
    float findx, findx2, fval;
    int indx, indx2;
    if (popnumber(x, &findx2) || popnumber(x, &findx) || popnumber(x, &fval))
        BAIL("'write1': bad arguments");
    if ((indx = findx) < 0 || indx >= MEMSIZE ||
        (indx2 = findx2) < 0 || indx2 >=  x->i_memory[indx].m_size)
            BAIL("'write1': variable number or index out of range");
    x->i_memory[indx].m_stuff[indx2].e_type = E_FLOAT;
    x->i_memory[indx].m_stuff[indx2].e_float = fval;
    return (0);
}

int f_plus(t_interp *x)         /* pop two numbers, add them, push result */
{                               /* This example is replicated for the other */
    float f1, f2;               /* binary operations using the macro below. */
    if (popnumber(x, &f2) || popnumber(x, &f1))
        BAIL("'+': bad or missing arguments");
    pushnumber(x, f1+f2);
    return (0);
}

#define BINOP(fname, string, expr) \
int fname(t_interp *x) \
{ \
    float f1, f2; \
    if (popnumber(x, &f2) || popnumber(x, &f1)) \
        BAIL(string ": bad or missing arguments"); \
    pushnumber(x, expr); \
    return (0); \
}

BINOP(f_minus, "-",   f1 - f2)
BINOP(f_times, "*",   f1 * f2)
BINOP(f_over,  "/",   (f2 != 0 ? f1/f2 : 0))
BINOP(f_gt,    ">",   f1 > f2)
BINOP(f_ge,    ">=",  f1 >= f2)
BINOP(f_lt,    "<",   f1 < f2)
BINOP(f_le,    "<=",  f1 <= f2)
BINOP(f_eq,    "==",  f1 == f2)
BINOP(f_ne,    "!=",  f1 != f2)
BINOP(f_max,   "max",  (f1 > f2 ? f1:f2))
BINOP(f_min,   "min",  (f1 < f2 ? f1:f2))
BINOP(f_div,   "div", \
    ((int)f2<0 ? ((int)f1-(int)f2+1)/(int)f2 : ((int)f2? (int)f1/(int)f2 : 0)))
BINOP(f_mod,   "mod", \
    ((int)f2<0 ? ((int)f1-(int)f2+1)%(int)f2 : ((int)f2? (int)f1%(int)f2 : 0)))

/**** the following are system functions -- the ones above are "pure". ****/
void elemtostring(t_elem *a, char *string)  /* utility for printouts */
{
    switch (a->e_type)
    {
    case E_FLOAT: sprintf(string, "%g", a->e_float); break;
    case E_FUNCTION: sprintf(string, "%s", 
        fntostring(a->e_fn)); break;
    case E_OPEN: sprintf(string, "("); break;
    case E_CLOSE: sprintf(string, ")"); break;
    }
}

int f_print(t_interp *x)        /* print one item from stack to stdout */
{
    return (RTN_OUTPUT);
}

int f_scan(t_interp *x)        /* input one number from stdin */
{
    return (RTN_INPUT);
}

int f_wait(t_interp *x)         /* wait F seconds */
{
    return (RTN_WAIT);
}

void dumpvec(t_elem *a, int size, char *name)   /* print out one list on the */
{                                               /* "console" - used below */
    int column = strlen(name) + 2, i;
    fprintf(stderr, "%s: ", name);
    for (i = 0; i < size; i++)
    {
        char str[80];
        elemtostring(&a[i], str);
        if ((column += strlen(str)) > 78)
        {
            fprintf(stderr, "\n");
            column = 0;
        }
        else if (i)
            fprintf(stderr, " ");
        fprintf(stderr, "%s", str);
        column += strlen(str);
    }
    fprintf(stderr, "\n");
}

int f_dump(t_interp *x)         /* print stack and all memory to stderr */
{
    int i, column;
    dumpvec(x->i_stack, x->i_stacksize, "stack");
    for (i = 0; i < MEMSIZE; i++)
        if (x->i_memory[i].m_size)
    {
        char str[80];
        sprintf(str, "%d", i);
        dumpvec(x->i_memory[i].m_stuff, x->i_memory[i].m_size, str);
    } 
    fprintf(stderr, "thank you.\n");
    return (0);
}

typedef struct _builtin_entry  /* Struct for searching for builtins by name */
{
    char *b_name;           /* the name */
    t_builtin b_fn;         /* the corresponding function */
} t_builtin_entry;

t_builtin_entry builtins[] = {  /* list of all built-ins - "pure" ones first */
{"if", f_if},           /* A F if - execute list A if number F nonzero */
{"while", f_while},     /* A B while - exec list A while list B returns !=0 */
{"r", f_read},          /* F r - read and return memory at location F */
{"w", f_write},         /* A F w - write block or number A to memory at F */
{"r1", f_read1},        /* F G r1 - read Gth element of Fth memory */
{"w1", f_write1},       /* F G H w1 - write H to Gth element of Fth memory */
{"+", f_plus},          /* F G +, etc - the usual binops */
{"-", f_minus},
{"*", f_times},
{"/", f_over},
{">", f_gt},            /* F G >, etc - logical binops return 0 or 1 */
{">=", f_ge},
{"<", f_lt},
{"<=", f_le},
{"==", f_eq},
{"!=", f_ne},
{"max", f_max},         /* F G max - greater of F and G */
{"min", f_min},         /* F G min - smaller of F and G */
{"div", f_div},         /* F G div - integer quotient */
{"mod", f_mod},         /* F G mod - integer modulus (remainder) */
{"print", f_print},     /* A print - print number or list A (system function) */
{"scan", f_scan},       /* F scan - scan a number to memory location F */
{"wait", f_wait},       /* F wait - sleep F seconds */
{"dump", f_dump},       /* dump - print stack and memory to console (system) */
};

t_builtin stringtofn(const char *s) /* look up a builtin function by name */
{
    int i;
    t_builtin_entry *e;
    for (i = 0; i < (int)(sizeof(builtins)/sizeof(*builtins)); i++)
        if (!strcmp(builtins[i].b_name, s))
            return (builtins[i].b_fn);
    return (0);
}

char *fntostring(t_builtin fn)      /* look string up for error printing */
{
    int i;
    t_builtin_entry *e;
    for (i = 0; i < (int)(sizeof(builtins)/sizeof(*builtins)); i++)
        if (builtins[i].b_fn == fn)
            return (builtins[i].b_name);
    return ("???");
}

#ifdef PD       /* implementation as a Pd object */
#include "m_pd.h"

static t_class *dum_class;
typedef struct _dum
{
    t_object x_obj;
    t_outlet *x_out1;
    t_outlet *x_out2;
    t_clock *x_clock;
    t_interp x_interp;
    int x_state;
    t_elem *x_block;
    int x_blocksize;
} t_dum;

static void dum_tick(t_dum *x);

static void *dum_new(t_float fnonrepeat)
{
    t_dum *x = (t_dum *)pd_new(dum_class);
    interp_init(&x->x_interp);
    x->x_out1 = outlet_new(&x->x_obj, &s_list);
    x->x_out2 = outlet_new(&x->x_obj, &s_float);
    x->x_clock = clock_new(x,  (t_method)dum_tick);
    x->x_state = 0;
    x->x_block = 0;
    x->x_blocksize = 0;
    return (x);
}

static void dum_clear(t_dum *x)
{
    interp_clear(&x->x_interp);
    clock_unset(x->x_clock);
}

static void dum_print(t_dum *x)
{
    f_dump(&x->x_interp);
}

#define MAXPRINT 100   /* LATER use malloc() instead */
static void dum_continue(t_dum *x)
{
    while (1)
    {
        x->x_state = interp_eval(&x->x_interp, x->x_block);
        if (x->x_state == RTN_WAIT)
        {
            float f;
            if (popnumber(&x->x_interp, &f))
            {
                post("'wait': bad or missing arguments");
                x->x_state = 0;
                return;
            }
            x->x_state = RTN_WAIT;
            clock_delay(x->x_clock, 1000*f);
            return;
        }
        else if (x->x_state == RTN_OUTPUT)
        {
            t_elem *block;
            int i;
            int blocksize;
            t_atom message[MAXPRINT];
            if (poplist(&x->x_interp, &block, &blocksize))
            {
                post("'print': bad or missing argument");
                x->x_state = 0;
                return;
            }
            if (blocksize > MAXPRINT)
            {
                post("'print': truncating to %d elements", MAXPRINT);
                blocksize = MAXPRINT;
            }
            for (i = 0; i < blocksize; i++)
            {
                if (block[i].e_type == E_FLOAT)
                    SETFLOAT(message+i, block[i].e_float);
                else if (block[i].e_type == E_OPEN)
                    SETSYMBOL(message+i, gensym("("));
                else if (block[i].e_type == E_CLOSE)
                    SETSYMBOL(message+i, gensym(")"));
                 else  SETSYMBOL(message+i, gensym(fntostring(block[i].e_fn)));
            }
            outlet_list(x->x_out1, 0, blocksize, message);
            x->x_state = RTN_DONE;
        }
        else
        {
            outlet_float(x->x_out2, x->x_state);
            return;
        }
    }
}

static void dum_tick(t_dum *x)
{
    if (x->x_state != RTN_WAIT)
    {
        bug("dum_tick");
        return;
    }
    dum_continue(x);
}

static void dum_exec(t_dum *x, t_symbol *s, int argc, t_atom *argv)
{
    x->x_block = realloc(x->x_block, sizeof(t_elem) * argc);
    x->x_blocksize = argc;
    int i, parendepth = 0;
    clock_unset(x->x_clock);
    for (i = 0; i < argc; i++)
    {
        if (argv[i].a_type == A_FLOAT)
        {
            x->x_block[i].e_type = E_FLOAT;
            x->x_block[i].e_float = argv[i].a_w.w_float;
        }
        else if (argv[i].a_type == A_SYMBOL)
        {
            char *str = argv[i].a_w.w_symbol->s_name;
            if (str[0] == '(' && !str[1])   /* open parenthesis */
            {
                x->x_block[i].e_type = E_OPEN;
                parendepth++;
            }
            else if (str[0] == ')' && !str[1])   /* close parenthesis */
            {
                if (--parendepth < 0)
                {
                    pd_error(x, "dum: too many close parentheses");
                    return;
                }
                x->x_block[i].e_type = E_CLOSE;
            }
            else if (x->x_block[i].e_fn = stringtofn(str))
                x->x_block[i].e_type = E_FUNCTION;
            else
            {
                pd_error(x, "dum: %s: unknown function:", str);
                return;
            }
        }
        else
        {
            pd_error(x, "dum: pointer or other bad element in list");
            return;
        }
    }
    if (parendepth)
    {
        pd_error(x, "dum: open parentheses not closed");
        return;
    }
    if (x->x_state != RTN_DONE)
        post("warning: dum interrupted in middle of execution by new exec");
    interp_start(&x->x_interp, argc);
    dum_continue(x);
}

static void dum_list(t_dum *x, t_symbol *s, int argc, t_atom *argv)
{
    int i;
    float f;
    if (x->x_state != RTN_INPUT)
        post("dum: warning: input dropped because not in 'read'");
    for (i = 0; i < argc; i++)
        pushnumber(&x->x_interp, atom_getfloatarg(i, argc, argv));
    x->x_state = RTN_DONE;
    dum_continue(x);
}

static void dum_free(t_dum *x)
{
    interp_free(&x->x_interp);
    clock_free(x->x_clock);
}

void dum_setup(void)
{
    dum_class = class_new(gensym("dum"), (t_newmethod)dum_new,
        (t_method)dum_free, sizeof(t_dum), 0, A_DEFFLOAT, 0);
    class_addmethod(dum_class, (t_method)dum_exec, gensym("exec"), A_GIMME, 0);
    class_addmethod(dum_class, (t_method)dum_clear, gensym("clear"), 0);
    class_addmethod(dum_class, (t_method)dum_print, gensym("print"), 0);
    class_addlist(dum_class, dum_list);
}

#else           /* PD not defined - by default we're a command line program */

t_interp maininterp;        /* one instance of the interpreter */

int main(int argc, char **argv)
{
    int nelem = 0, parendepth = 0;
    t_elem *input = 0;
    while (1)
    {
        char string[81];
        if (scanf("%80s", string) < 1)
            exit(0);
        input = (t_elem *)realloc(input, (nelem+1)*sizeof(t_elem));
        if (string[0] == '(')
        {
            input[nelem].e_type = E_OPEN;
            parendepth++;
            if (string[1])
                fprintf(stderr, "warning: characters after '(' dropped\n");
        }
        else if (string[0] == ')')
        {
            if (parendepth <= 0)
                goto drop;
            input[nelem].e_type = E_CLOSE;
            parendepth--;
            if (string[1])
                fprintf(stderr, "warning: characters after ')' dropped\n");
        }
        else if (isdigit(string[0]) || string[0] == '.' || 
            (string[0] == '-' && string[1]))
        {
            input[nelem].e_type = E_FLOAT;
            input[nelem].e_float = atof(string);
        }
        else if (input[nelem].e_fn = stringtofn(string))
            input[nelem].e_type = E_FUNCTION;
        else
        {
            fprintf(stderr, "%s: unknown function\n", string);
            goto drop;
        }
        nelem++;
        if (!parendepth)
        {
            interp_start(&maininterp, nelem);
            while (1)
            {
                int rtn = interp_eval(&maininterp, input);
                if (rtn == RTN_DONE)
                    break;
                else if (rtn == RTN_ERROR)
                    break;
                else if (rtn == RTN_INPUT)
                {
                    float f;
                    fprintf(stderr, "input: ");
                    fflush(stderr);
                    if (scanf("%f", &f) < 1)
                    {
                        fprintf(stderr, "scan failed\n");
                        break;
                    }
                    if (pushnumber(&maininterp, f))
                        break;
                }
                else if (rtn == RTN_OUTPUT)
                {
                    t_elem *block;
                    int i;
                    int blocksize;
                    if (poplist(&maininterp, &block, &blocksize))
                        BAIL("'pop': empty stack");
                    for (i = 0; i < blocksize; i++)
                    {
                        char str[80];
                        elemtostring(&block[i], str);
                        printf("%s ", str);
                    }
                    printf("\n");
                    fflush(stdout);
                }
                else if (rtn == RTN_WAIT)
                {
                    float f;
                    struct timeval timout;
                    if (popnumber(&maininterp, &f))
                        BAIL("'wait': bad or missing arguments");
                    timout.tv_sec = 0;
                    timout.tv_usec = (f > 0 ? f*1000000. : 0);
                    select(0, 0, 0, 0, &timout);
                }
            }
            free(input);
            input = 0;
            nelem = 0;
        }
        continue;
    drop: ;
        input = (t_elem *)realloc(input, nelem*sizeof(t_elem));
    }
}

#endif  /* end of #ifdef PD */
